/**
 ** ivmobTrackTaskCodels.c
 **
 ** Codels called by execution task ivmobTrackTask
 **
 ** Author:
 ** Date:
 **
 **/
#include <math.h>
#include <stdint.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "server/ivmobHeader.h"
#include "ivmobConst.h"

#include <ivmob/ivmob.h>

static GENPOS_CART_REF stillRef;

/*------------------------------------------------------------------------
 * TrackSpeed
 *
 * Description:
 *
 * Reports:      OK
 *              S_ivmob_POSTER_NOT_FOUND
 *              S_ivmob_BAD_REF
 *              S_ivmob_CMD_STOP_TRACK
 */

/* findTrackedPoster  -  codel START of TrackSpeed
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
findTrackedPoster(IVMOB_TRACK_PARAMS *params, int *report)
{
	char *trackedPoster = (char *)params->posterName.name;
	POSTER_ID posterTrackedId;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	IVMOB_DYNAMIC_STR *dynamics = &SDI_F->dynamics;

	/* Poster Identifier */
	if (posterFind((char *) trackedPoster, &posterTrackedId) != OK)  {
		*report = S_ivmob_POSTER_NOT_FOUND;
		return(END);
	}
	ivmobDataStrId->trackStr.posterId = (long) posterTrackedId;
	ivmobDataStrId->status.rs_mode = IVMOB_TRACK;
	/* Set stillRef to current position of the robot */
	stillRef.x = robot->xRef;
	stillRef.y = robot->yRef;
	stillRef.theta = robot->theta;
	stillRef.v = 0.0;
	stillRef.w = 0.0;
	stillRef.vmax = dynamics->vMax;
	stillRef.wmax = dynamics->wMax;
	stillRef.linAccelMax = dynamics->linAccelMax;
	stillRef.angAccelMax = dynamics->angAccelMax;
	switch (params->mode) {
	case GENPOS_TRACK_CONFIG:
		ivmobDataStrId->ref.dataType = GENPOS_POS_AND_SPEED_DATA;
		stillRef.dataType = GENPOS_POS_AND_SPEED_DATA;
		break;
	case GENPOS_TRACK_SPEED:
		ivmobDataStrId->ref.dataType = GENPOS_SPEED_DATA;
		stillRef.dataType = GENPOS_SPEED_DATA;
		break;
	default:		
		*report = S_ivmob_BAD_REF;
		return FAIL;
	}
	stillRef.numRef = -1;

	printf("ivmob tracking on %s\n", (char *)trackedPoster);
	return(EXEC);
}


static ACTIVITY_EVENT
pumpConfigReference(int *report)
{
	GENPOS_CART_REF ivmobOrders;
	GENPOS_CART_REF *refId = &ivmobDataStrId->ref;
	GENPOS_CART_STATE *configId = &ivmobDataStrId->robot;
	IVMOB_DYNAMIC_STR *dynamics = &ivmobDataStrId->dynamics;
	static int prevRef = -2;

	/* Lire les consignes sur le poster externe */
	if (posterRead ((POSTER_ID)ivmobDataStrId->trackStr.posterId, 0,
		(void *) &ivmobOrders,
		sizeof (GENPOS_CART_REF)) == sizeof (GENPOS_CART_REF)) {

		/* Check validity of that  */
		if (ivmobOrders.numRef == -1) {
			if (prevRef != -1) {
				printf("Pilo numRef = -1\n");
			}
			prevRef = -1;
			/* Not a valid reference don't move */
			memcpy(refId, &stillRef, sizeof(GENPOS_CART_REF));
			return (EXEC);
		}
		/* Limit speeds and accelerations */
		/* XXX may be set an error */
		if (ivmobOrders.vmax > dynamics->vMax)
			ivmobOrders.vmax = dynamics->vMax;
		if (ivmobOrders.wmax > dynamics->wMax)
			ivmobOrders.wmax = dynamics->wMax;
		if (ivmobOrders.linAccelMax > dynamics->linAccelMax)
			ivmobOrders.linAccelMax = dynamics->linAccelMax;
		if (ivmobOrders.angAccelMax > dynamics->angAccelMax)
			ivmobOrders.angAccelMax = dynamics->angAccelMax;

		/* Send reference */
		refId->backFlag = (ivmobOrders.v > 0 ?
				   GENPOS_FORWARD_MOTION :
				   GENPOS_BACKWARD_MOTION);
		refId->v = fabs(ivmobOrders.v);
		refId->vt = 0;
		refId->w = ivmobOrders.w;
		refId->vmax = ivmobOrders.vmax;
		refId->wmax = ivmobOrders.wmax;
		refId->linAccelMax = ivmobOrders.linAccelMax;
		refId->angAccelMax = ivmobOrders.angAccelMax;
		refId->x = ivmobOrders.x;
		refId->y = ivmobOrders.y;
		refId->theta =  angleLimit(ivmobOrders.theta);
		refId->dataType = GENPOS_POS_AND_SPEED_DATA;
		/* store position in case of stop */
		stillRef.x = refId->x;
		stillRef.y = refId->y;
		stillRef.theta = refId->theta;
		prevRef = ivmobOrders.numRef;
		return(EXEC);
	}
	*report = errnoGet();
	return(END);
}

static ACTIVITY_EVENT
pumpSpeedReference(int *report)
{
	GENPOS_CART_SPEED ivmobOrders;
	GENPOS_CART_REF *refId = &ivmobDataStrId->ref;
	GENPOS_CART_STATE *configId = &ivmobDataStrId->robot;

	/* Lire les consignes sur le poster externe */
	if (posterRead ((POSTER_ID)ivmobDataStrId->trackStr.posterId, 0,
		(void *) &ivmobOrders, sizeof (GENPOS_CART_SPEED)) 
	    == sizeof (GENPOS_CART_SPEED)) {

		/* Transmettre la consigne */
		refId->backFlag = (ivmobOrders.v > 0 ?
		    GENPOS_FORWARD_MOTION : GENPOS_BACKWARD_MOTION);
		refId->v = fabs(ivmobOrders.v);
		/* printf("pumpSpeedRef: %lf\n", refId->v); */
		refId->vt = 0;
		refId->w = ivmobOrders.w;
		refId->vmax = ivmobOrders.vmax;
		refId->wmax = ivmobOrders.wmax;
		refId->linAccelMax = ivmobOrders.linAccelMax;
		refId->angAccelMax = ivmobOrders.angAccelMax;
		refId->dataType = GENPOS_SPEED_DATA;
		return(EXEC);
	} else {
		printf("pumpSpeedReference: posterRead error\n");
	}
	*report = errnoGet();
	return(END);
}

/* pumpReference  -  codel EXEC of TrackSpeed
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pumpReference(IVMOB_TRACK_PARAMS *params, int *report)
{
	GENPOS_TRACK *track = &ivmobDataStrId->trackStr;
	IVMOB_STATUS_STR *status = &ivmobDataStrId->status;

	/* Check if mode changed */
	switch (status->rs_mode) {
	case IVMOB_MOTORS_OFF:
		*report = S_ivmob_MOTORS_OFF;
		return END;
	case IVMOB_EMERGENCY:
		*report = S_ivmob_EMERGENCY_STOP;
		return END;
	}

	switch (track->trackMode) {
	case GENPOS_TRACK_POS:
		return pumpConfigReference(report);
	case GENPOS_TRACK_CONFIG:
		return pumpConfigReference(report);
	case GENPOS_TRACK_SPEED:
		return pumpSpeedReference(report);
	default:
		*report = EXEC_TASK_BILAN(IVMOB_MOTIONTASK_NUM);
		return(END);

	}
}

/* smoothStopTrack  -  codel END of TrackSpeed
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
smoothStopTrack(IVMOB_TRACK_PARAMS *params, int *report)
{
	GENPOS_CART_REF *refId;	       /* Id structure de consignes */
	GENPOS_CART_STATE *robotId;	/* Id config du robot */
	GENPOS_TRACK *trackId;
	IVMOB_DYNAMIC_STR *dynamics;
	double dtArret;                /* Temps d'arret */
	double dtRetard;               /* Temps de prise en compte de la consigne */
	double dsArret;                /* Longueur d'arret */
	double dtLoco;                 /* Periode d'asserv de la loco */
	double angle;                  /* Moyenne entre angles d'init et d'arret */
	double theta;

	printf("ivmob smoothStopTrack\n");

	/* Get pointers to IDS members */
	robotId = &(ivmobDataStrId->robot);
	refId = &ivmobDataStrId->ref;
	trackId = &ivmobDataStrId->trackStr;
	dynamics = &ivmobDataStrId->dynamics;

	if (trackId->trackMode == GENPOS_TRACK_SPEED) {
		/* Set a null speed and stop the tracking */
		trackId->trackMode = GENPOS_NO_TRACKING;
		refId->dataType = GENPOS_SPEED_DATA;
		refId->x = robotId->xRob;
		refId->y = robotId->yRob;
		refId->theta = robotId->theta;
		refId->v = 0.0;
		refId->w = 0.0;
	} else {
		refId->dataType = GENPOS_POS_AND_SPEED_DATA;

		/* Determination de la periode d'asserv de la loco */
		dtLoco = EXEC_TASK_PERIOD(IVMOB_TRACKTASK_NUM);

		/* Temps d'arret minimal (tps de decel) */
		dtArret = MAX (fabs(robotId->v) / dynamics->linAccelMax,
			fabs(robotId->w) / dynamics->angAccelMax);

		/* Temps de prise en compte (v = cst) */
		dtRetard =  1 * dtLoco;

		/* Angle final (avec periode de prise en compte) */
		theta = robotId->theta + robotId->w * (dtArret / 2 + dtRetard);

		/* Angle moyen */
		angle = (robotId->theta + theta) / 2;

		/* Longueur d'arret */
		dsArret = robotId->v * (dtArret / 2 + dtRetard);

		/* Prevision position du centre */
		refId->x = robotId->xRob + dsArret * cos (angle);
		refId->y = robotId->yRob + dsArret * sin (angle);

		/* Vitesses de consigne nulles */
		refId->v = 0. ;
		refId->w = 0. ;

		/* Indiquer l'arret du tracking */
	}
	trackId->trackMode = GENPOS_NO_TRACKING;
	ivmobDataStrId->status.rs_mode = IVMOB_IDLE;
	posterWrite(IVMOB_STATUS_POSTER_ID, 0, &(ivmobDataStrId->status), sizeof(IVMOB_STATUS_STR));
	return ETHER;
}
